#!/usr/bin/env perl

use XyceRegression::Tools;
use MeasureCommon;

$Tools = XyceRegression::Tools->new();
#$Tools->setDebug(1);
#$Tools->setVerbose(1);

# The input arguments to this script are: 
# $1 = location of Xyce binary
# $2 = location of xyce_verify.pl script
# $3 = location of compare script 
# $4 = location of circuit file to test
# $5 = location of gold standard prn file

# If Xyce does not produce a prn file, then we return exit code 10.
# If Xyce succeeds, but the test fails, then we return exit code 2.
# If the shell script fails for some reason, then we return exit code 1.
# Otherwise we return the exit code of compare or xyce_verify.pl

# Since the shell script runs Xyce and the comparison program, it is
# responsible for capturing any error output from Xyce and the STDOUT & STDERR
# from the comparison program.  The outside script, run_xyce_regression,
# expects the STDERR output from Xyce to go into $CIRFILE.err, the STDOUT
# output from comparison to go into $CIRFILE.prn.out and the STDERR output from
# comparison to go into $CIRFILE.prn.err.  

use Getopt::Long;
use Scalar::Util qw(looks_like_number);

&GetOptions( "verbose!" => \$verbose );
$XYCE=$ARGV[0];
$XYCE_VERIFY=$ARGV[1];
#$XYCE_COMPARE=$ARGV[2];
$CIRFILE=$ARGV[3];
#$GOLDPRN=$ARGV[4];

# these are the tolerances used in comparing the measures calculated by this 
# script to those calculated by Xyce.  Zerotol is needed because some measured
# values should be zero (and reltol won't make sense in that case).

# phase is output in degrees and for fourier components with very small magnitude, 
# it can be have several degrees of scatter.  Thus it gets its own abstol
my $phaseAbsTol = 1.0;
my $relTol = 0.01;
my $zeroTol = 1.0e-8;
my $expectedDoubleCount=51;
my $defaultPrecision=6;

if (defined($verbose)) { $Tools->setVerbose(1); }

sub verbosePrint { $Tools->verbosePrint(@_); }

# run Xyce and check for the proper files
$CMD="$XYCE $CIRFILE > $CIRFILE.out";
$retval=system($CMD)>>8;
if ($retval != 0) { print "Exit code = $retval\n"; exit $retval; }
if (not -s "$CIRFILE.csd" ) { print "Exit code = 14\n"; exit 14; }

#
# Did we make a measure file
#
if (not -s "$CIRFILE.mt0" ) { print "Exit code = 17\n"; exit 17; }

# do re-measure
use File::Copy;
move("$CIRFILE.mt0","$CIRFILE.temp.mt0");
move("$CIRFILE.mt1","$CIRFILE.temp.mt1");
move("$CIRFILE.mt2","$CIRFILE.temp.mt2");

# here is the command to run xyce with remeasure
my $CMD="$XYCE -remeasure $CIRFILE.csd $CIRFILE > $CIRFILE.remeasure.out";
$retval=system($CMD)>>8;
  
if ($retval != 0) { print "Exit code = $retval\n"; exit $retval; }
if (not -s "$CIRFILE.csd" ) { print "Exit code = 14\n"; exit 14; }

# Did we make a measure file
if (not -s "$CIRFILE.mt0" ) { print "Exit code = 17\n"; exit 17; }

# rename the files
move("$CIRFILE.mt0","$CIRFILE.remeasure.mt0");
move("$CIRFILE.temp.mt0","$CIRFILE.mt0");
move("$CIRFILE.mt1","$CIRFILE.remeasure.mt1");
move("$CIRFILE.temp.mt1","$CIRFILE.mt1");
move("$CIRFILE.mt2","$CIRFILE.remeasure.mt2");
move("$CIRFILE.temp.mt2","$CIRFILE.mt2");

# now compare the re-measured and original file contents
$retval = MeasureCommon::compareFourierMeasureFiles("$CIRFILE.remeasure.mt0", "$CIRFILE.mt0", $phaseAbsTol, $relTol, $zeroTol, $defaultPrecision,$expectedDoubleCount);
if ( $retval != 0 )
{
  verbosePrint "test Failed for re-measure of Step 0!\n";
  print "Exit code = $retval\n"; 
  exit $retval;
}

$retval =  MeasureCommon::compareFourierMeasureFiles("$CIRFILE.remeasure.mt1","$CIRFILE.mt1", $phaseAbsTol, $relTol, $zeroTol, $defaultPrecision,$expectedDoubleCount);
if ( $retval != 0 )
{
  verbosePrint "test Failed for re-measure of Step 1!\n";
  print "Exit code = $retval\n"; 
  exit $retval;
}
  
$retval =  MeasureCommon::compareFourierMeasureFiles("$CIRFILE.remeasure.mt2", "$CIRFILE.mt2", $phaseAbsTol, $relTol, $zeroTol, $defaultPrecision,$expectedDoubleCount);
if ( $retval != 0 )
{
  verbosePrint "test Failed for re-measure of Step 2!\n";
  print "Exit code = $retval\n"; 
  exit $retval;
}
  
verbosePrint "test Passed!\n";
print "Exit code = $retval\n"; 
exit $retval;


